#!/usr/bin/env python
# coding: utf-8

# In[1]:


import at
import at.plot
import numpy as np
import matplotlib.pyplot as aplt

ring = at.load_mat('./dba.mat', mat_key='RING')
arc = at.load_mat('./arc.mat', mat_key='ARC')


# In[2]:


ring.plot_beta()
arc.plot_beta()


# There are many other handy functions for modifying the lattice in this way including a lot of the standard python list modifications:
# > 1. sbreak
# > 2. slice
# > 3. insert
# > 4. remove
# > 5. append
# > 6. pop
# > 7. repeat
# > 8. reduce 
# 

# ## sbreak

# In[3]:


get_ipython().run_line_magic('pinfo', 'ring.sbreak')


# In[4]:


ring_tmp = ring.deepcopy()
ring_tmp[0]


# In[5]:


ring_tmp = ring_tmp.sbreak(0.2, at.Marker('hello'))
print(ring_tmp[0],'\n', ring_tmp[1],'\n', ring_tmp[2])


# In[6]:


ring_tmp = ring_tmp.sbreak(2.8, at.Marker('QuadSplit'))
print(ring_tmp[3],'\n', ring_tmp[4],'\n', ring_tmp[5])


# For quadrupoles, there is an flag for the inclusion or exclusion of fringe fields at the entrance and exit of the quadrupoles. When you split the quadrupole, the internal fringe fields are disabled, and the outer fringe fields are kept. Similarly when you split a dipole, the exit and entrance angles are modified to ensure a continuous bend. 

# ## slice / divide / merge

# You slice a lattice, you divide an element

# In[7]:


ring_tmp = ring.deepcopy()


# In[8]:


quad = ring_tmp[3]
print(quad, '\n')
quad_div = quad.divide([0.5, 0.4])
print(quad_div[0], '\n', quad_div[1])


# It may be useful to make nice smooth plots (for example of optics functions in the middle of elements and not just at the edges). For this you can use slice.

# In[9]:


get_ipython().run_line_magic('pinfo', 'ring_tmp.slice')


# In[10]:


new_ring_tmp = ring_tmp.slice(slices=2000)


# In[11]:


msk = at.get_value_refpts(ring_tmp, np.arange(len(ring_tmp)), 'Length') == 0
print(len(ring_tmp) - np.sum(msk))

msk = at.get_value_refpts(new_ring_tmp, np.arange(len(new_ring_tmp)), 'Length') == 0
print(len(new_ring_tmp) - np.sum(msk))


# Zero length elements are of course not touched, but one can also specify a dz too

# We can also do the opposite of split, which is to merge two identical and adjacent elements

# In[12]:


q = ring_tmp[3]


# In[13]:


print(q)


# In[14]:


q.merge(q)


# In[15]:


print(q)


# In[16]:


dr = ring_tmp[0]
print(dr)


# In[17]:


dr.merge(dr)
print(dr)


# ## insert / append / pop / remove

# Lattice objects are fundamentally lists. Which means that all of the standard python operations that can be performed on lists are valid for a lattice object

# In[18]:


ring_tmp = ring.deepcopy()


# In[19]:


ring_tmp.insert(4, at.Marker('insert_test'))


# In[20]:


print(ring_tmp[4])


# Append puts elements automatically at the end. It is equivalent to insert with an index of -1
# 

# In[21]:


ring_tmp.append(at.Marker('append_test'))
print(ring_tmp[-1])


# Pop removes and returns the element at the specified index

# In[22]:


mark = ring_tmp.pop(-1)


# In[23]:


print(mark)
print(ring_tmp[-1])


# In[24]:


mrk = at.Marker('ahoyhoy')
ring_tmp.append(mrk)
ring_tmp.remove(mrk)


# In[25]:


at.get_uint32_index(ring_tmp, 'ahoy*')


# ## repeat

# In[26]:


arc_tmp = arc.deepcopy()


# In[27]:


get_ipython().run_line_magic('pinfo', 'arc_tmp.repeat')


# In[28]:


arc_tmp.periodicity


# In[29]:


ring_tmp = arc_tmp.repeat(arc_tmp.periodicity)


# In[30]:


ring_tmp.periodicity
print(len(arc_tmp), len(ring_tmp), 33*25)


# ## reduce

# Very important and useful function! This function will remove all irrelevant elements from your lattice, and merge compatible adjacent elements. For example, say you have split a quadrupole in the center and put a marker there. The marker has an 'IdentityPass', pass method which means it will be removed. The now adjacent quadrupoles will be merged. Very useful for reducing as much as possible the number of elements that need to be actively tracked in your lattice.

# In[31]:


ring_tmp = ring.deepcopy()


# In[32]:


get_ipython().run_line_magic('pinfo', 'ring_tmp.reduce')


# In[33]:


len(ring_tmp)


# In[34]:


ring_tmp = ring_tmp.reduce()


# In[35]:


len(ring_tmp)


# ### masks / refpts

# In[36]:


ring_tmp = ring.deepcopy()
len(ring)


# You have a ring. It has for example, 876 elements. There are a variety of different ways to select the desired elements in the list.

# In[37]:


msk = np.zeros(len(ring_tmp), dtype=bool)
msk[10:20] = True
ring_tmp[msk]


# In[38]:


args = np.array([1,2,3,5,6,234,450])
ring_tmp[args]


# As we have already seen. We have the function 'get_uint32_index' to get the indexes of elements matching a certain criteria. We also have 'get_bool_index' in order to obtain a mask for the same values.

# In[39]:


msk = at.get_bool_index(ring,'BPM*')
inds = at.get_uint32_index(ring,'BPM*')


# In[40]:


print(np.where(msk)[0])
print(inds)


# In[ ]:




